@@ -26,6 +26,10 @@
 #include "neatvnc.h"
 #include "config.h"
 
+#ifdef HAVE_CRYPTO
+#include "crypto.h"
+#endif
+
 #ifdef ENABLE_TLS
 #include <gnutls/gnutls.h>
 #endif
@@ -43,6 +47,13 @@ enum nvnc_client_state {
 	VNC_CLIENT_STATE_WAITING_FOR_VENCRYPT_VERSION,
 	VNC_CLIENT_STATE_WAITING_FOR_VENCRYPT_SUBTYPE,
 	VNC_CLIENT_STATE_WAITING_FOR_VENCRYPT_PLAIN_AUTH,
+#endif
+#ifdef HAVE_CRYPTO
+	VNC_CLIENT_STATE_WAITING_FOR_APPLE_DH_RESPONSE,
+	VNC_CLIENT_STATE_WAITING_FOR_RSA_AES_PUBLIC_KEY,
+	VNC_CLIENT_STATE_WAITING_FOR_RSA_AES_CHALLENGE,
+	VNC_CLIENT_STATE_WAITING_FOR_RSA_AES_CLIENT_HASH,
+	VNC_CLIENT_STATE_WAITING_FOR_RSA_AES_CREDENTIALS,
 #endif
 	VNC_CLIENT_STATE_WAITING_FOR_INIT,
 	VNC_CLIENT_STATE_READY,
@@ -53,6 +64,9 @@ struct stream;
 struct aml_handler;
 struct aml_idle;
 struct nvnc_display;
+struct crypto_key;
+struct crypto_rsa_pub_key;
+struct crypto_rsa_priv_key;
 
 struct nvnc_common {
 	void* userdata;
@@ -69,6 +83,8 @@ struct nvnc_client {
 	struct nvnc_common common;
 	int ref;
 	struct stream* net_stream;
+	char hostname[256];
+	char username[256];
 	struct nvnc* server;
 	enum nvnc_client_state state;
 	bool has_pixfmt;
@@ -87,16 +103,36 @@ struct nvnc_client {
 	uint32_t known_width;
 	uint32_t known_height;
 	struct cut_text cut_text;
-	bool is_qemu_key_ext_notified;
+	bool is_ext_notified;
 	struct encoder* encoder;
 	uint32_t cursor_seq;
+	int quality;
+
+#ifdef HAVE_CRYPTO
+	struct crypto_key* apple_dh_secret;
+
+	struct {
+		enum crypto_hash_type hash_type;
+		enum crypto_cipher_type cipher_type;
+		size_t challenge_len;
+		uint8_t challenge[32];
+		struct crypto_rsa_pub_key* pub;
+	} rsa;
+#endif
 };
 
 LIST_HEAD(nvnc_client_list, nvnc_client);
 
+enum nvnc__socket_type {
+	NVNC__SOCKET_TCP,
+	NVNC__SOCKET_UNIX,
+	NVNC__SOCKET_WEBSOCKET,
+};
+
 struct nvnc {
 	struct nvnc_common common;
 	int fd;
+	enum nvnc__socket_type socket_type;
 	struct aml_handler* poll_handle;
 	struct nvnc_client_list clients;
 	char name[256];
@@ -107,6 +143,7 @@ struct nvnc {
 	nvnc_fb_req_fn fb_req_fn;
 	nvnc_client_fn new_client_fn;
 	nvnc_cut_text_fn cut_text_fn;
+	nvnc_desktop_layout_fn desktop_layout_fn;
 	struct nvnc_display* display;
 	struct {
 		struct nvnc_fb* buffer;
@@ -115,10 +152,17 @@ struct nvnc {
 	} cursor;
 	uint32_t cursor_seq;
 
-#ifdef ENABLE_TLS
-	gnutls_certificate_credentials_t tls_creds;
+	enum nvnc_auth_flags auth_flags;
 	nvnc_auth_fn auth_fn;
 	void* auth_ud;
+
+#ifdef ENABLE_TLS
+	gnutls_certificate_credentials_t tls_creds;
+#endif
+
+#ifdef HAVE_CRYPTO
+	struct crypto_rsa_pub_key* rsa_pub;
+	struct crypto_rsa_priv_key* rsa_priv;
 #endif
 
 	uint32_t n_damage_clients;
