@@ -29,8 +29,11 @@ enum rfb_security_type {
 	RFB_SECURITY_TYPE_INVALID = 0,
 	RFB_SECURITY_TYPE_NONE = 1,
 	RFB_SECURITY_TYPE_VNC_AUTH = 2,
+	RFB_SECURITY_TYPE_RSA_AES = 5,
 	RFB_SECURITY_TYPE_TIGHT = 16,
 	RFB_SECURITY_TYPE_VENCRYPT = 19,
+	RFB_SECURITY_TYPE_APPLE_DH = 30,
+	RFB_SECURITY_TYPE_RSA_AES256 = 129,
 };
 
 enum rfb_security_handshake_result {
@@ -45,6 +48,8 @@ enum rfb_client_to_server_msg_type {
 	RFB_CLIENT_TO_SERVER_KEY_EVENT = 4,
 	RFB_CLIENT_TO_SERVER_POINTER_EVENT = 5,
 	RFB_CLIENT_TO_SERVER_CLIENT_CUT_TEXT = 6,
+	RFB_CLIENT_TO_SERVER_NTP = 160,
+	RFB_CLIENT_TO_SERVER_SET_DESKTOP_SIZE = 251,
 	RFB_CLIENT_TO_SERVER_QEMU = 255,
 };
 
@@ -63,17 +68,21 @@ enum rfb_encodings {
 	RFB_ENCODING_OPEN_H264 = 50,
 	RFB_ENCODING_CURSOR = -239,
 	RFB_ENCODING_DESKTOPSIZE = -223,
-	RFB_ENCODING_JPEG_HIGHQ = -23,
-	RFB_ENCODING_JPEG_LOWQ = -32,
 	RFB_ENCODING_QEMU_EXT_KEY_EVENT = -258,
+	RFB_ENCODING_EXTENDEDDESKTOPSIZE = -308,
 	RFB_ENCODING_PTS = -1000,
+	RFB_ENCODING_NTP = -1001,
 };
 
+#define RFB_ENCODING_JPEG_HIGHQ -23
+#define RFB_ENCODING_JPEG_LOWQ -32
+
 enum rfb_server_to_client_msg_type {
 	RFB_SERVER_TO_CLIENT_FRAMEBUFFER_UPDATE = 0,
 	RFB_SERVER_TO_CLIENT_SET_COLOUR_MAP_ENTRIES = 1,
 	RFB_SERVER_TO_CLIENT_BELL = 2,
 	RFB_SERVER_TO_CLIENT_SERVER_CUT_TEXT = 3,
+	RFB_SERVER_TO_CLIENT_NTP = 160,
 };
 
 enum rfb_vencrypt_subtype {
@@ -86,9 +95,28 @@ enum rfb_vencrypt_subtype {
 	RFB_VENCRYPT_X509_PLAIN,
 };
 
+enum rfb_resize_initiator {
+	RFB_RESIZE_INITIATOR_SERVER = 0,
+	RFB_RESIZE_INITIATOR_THIS_CLIENT = 1,
+	RFB_RESIZE_INITIATOR_OTHER_CLIENT = 2,
+};
+
+enum rfb_resize_status {
+	RFB_RESIZE_STATUS_SUCCESS = 0,
+	RFB_RESIZE_STATUS_PROHIBITED = 1,
+	RFB_RESIZE_STATUS_OUT_OF_RESOURCES = 2,
+	RFB_RESIZE_STATUS_INVALID_LAYOUT = 3,
+	RFB_RESIZE_STATUS_REQUEST_FORWARDED = 4,
+};
+
+enum rfb_rsa_aes_cred_subtype {
+	RFB_RSA_AES_CRED_SUBTYPE_USER_AND_PASS = 1,
+	RFB_RSA_AES_CRED_SUBTYPE_ONLY_PASS = 2,
+};
+
 struct rfb_security_types_msg {
 	uint8_t n;
-	uint8_t types[1];
+	uint8_t types[0];
 } RFB_PACKED;
 
 struct rfb_error_reason {
@@ -171,6 +199,25 @@ struct rfb_server_fb_rect {
 	int32_t encoding;
 } RFB_PACKED;
 
+struct rfb_screen {
+	uint32_t id;
+	uint16_t x;
+	uint16_t y;
+	uint16_t width;
+	uint16_t height;
+	uint32_t flags;
+} RFB_PACKED;
+
+struct rfb_client_set_desktop_size_event_msg {
+	uint8_t type;
+	uint8_t padding;
+	uint16_t width;
+	uint16_t height;
+	uint8_t number_of_screens;
+	uint8_t padding2;
+	struct rfb_screen screens[0];
+} RFB_PACKED;
+
 struct rfb_server_fb_update_msg {
 	uint8_t type;
 	uint8_t padding;
@@ -192,3 +239,30 @@ struct rfb_vencrypt_plain_auth_msg {
 	uint32_t password_len;
 	char text[0];
 } RFB_PACKED;
+
+struct rfb_ntp_msg {
+	uint8_t type;
+	uint8_t padding[3];
+	uint32_t t0, t1, t2, t3;
+} RFB_PACKED;
+
+struct rfb_apple_dh_server_msg {
+	uint16_t generator;
+	uint16_t key_size;
+	uint8_t modulus_and_key[0];
+} RFB_PACKED;
+
+struct rfb_apple_dh_client_msg {
+	uint8_t encrypted_credentials[128];
+	uint8_t public_key[0];
+} RFB_PACKED;
+
+struct rfb_rsa_aes_pub_key_msg {
+	uint32_t length;
+	uint8_t modulus_and_exponent[0];
+} RFB_PACKED;
+
+struct rfb_rsa_aes_challenge_msg {
+	uint16_t length;
+	uint8_t challenge[0];
+} RFB_PACKED;
