@@ -46,11 +46,6 @@ static void nvnc_display__on_resampler_done(struct nvnc_fb* fb,
 
 	assert(self->server);
 
-	struct nvnc_client* client;
-	LIST_FOREACH(client, &self->server->clients, link)
-		if (client->encoder)
-			encoder_push(client->encoder, fb, damage);
-
 	// TODO: Shift according to display position
 	nvnc__damage_region(self->server, damage);
 }
